unit GLPicture;

interface

uses
	Windows, SysUtils, Classes, Controls, Graphics, OpenGL, Math;

type
  TSinglePoint = packed record
  	X, Y: Single;
  end;

  TSingleRect = packed record
  	case Integer of
	  	0:(Left, Bottom, Right, Top: Single);
      1:(P1, P2: TSinglePoint);
  end;

  TGLRect = class(TPersistent)
  protected
    FRect: TSingleRect;
    function GetHeight: Single;
    function GetLeft: Single;
    function GetRight: Single;
    function GetTop: Single;
    function GetWidth: Single;
  public
  	constructor Create;
    property Rect: TSingleRect read FRect write FRect;
    property Left: Single read GetLeft;
    property Bottom: Single read GetLeft;
    property Right: Single read GetRight;
    property Top: Single read GetTop;
    property Width: Single read GetWidth;
    property Height: Single read GetHeight;
  end;

  TGLScale = class(TPersistent)
  private
    function GetHeight: Single;
    function GetLeft: Single;
    function GetRect: TSingleRect;
    function GetRight: Single;
    function GetTop: Single;
    function GetWidth: Single;
  protected
  	FRect: TGLRect;
  public
  	constructor Create;
    destructor Destroy; override;
    procedure Change(Left, Bottom, Right, Top: Single); overload;
    procedure Change(P1, P2: TSinglePoint); overload;
    procedure Change(Rect: TSingleRect); overload;
    property Rect: TSingleRect read GetRect;
    property Left: Single read GetLeft;
    property Bottom: Single read GetLeft;
    property Right: Single read GetRight;
    property Top: Single read GetTop;
    property Width: Single read GetWidth;
    property Height: Single read GetHeight;
  end;

	TGLBasePicture = class(TCustomControl)
  private
    FDC: HDC;
    FRC: HGLRC;
    FPalette: HPalette;
  protected
  	procedure InitVCL();
  	procedure InitGL();
    procedure InitGL2();
    procedure InitDC(DC: HDC; Palette: HPalette);
    procedure MakeCurrent();
    procedure PageFlip();
    procedure RedoPalette();
    procedure CleanupGL();
    procedure Resize; override;
    procedure Paint; override;
	  property DC: HDC read FDC;
    property RC: HGLRC read FRC;
    property Palette: HPalette read FPalette;
  public
  	constructor Create(AOWner: TComponent); override;
    destructor Destroy; override;
    procedure DRAW(); virtual;
    property Color;
  end;

	TGLPicture = class(TGLBasePicture)
  protected
    FScale: TGLScale;
  public
  	constructor Create(AOWner: TComponent); override;
    destructor Destroy; override;
    procedure Paint(); override;
  	procedure Clear();
    property Scale: TGLScale read FScale;
  end;

function SingleRect(Left, Bottom, Right, Top: Single): TSingleRect; overload;
function SingleRect(P1, P2: TSinglePoint): TSingleRect; overload;

var
	SingleBuffer: Boolean = True;
  ColorIndex: Boolean = False;

implementation

function SingleRect(Left, Bottom, Right, Top: Single): TSingleRect;
begin
	Result.Left := Min(Left, Right);
  Result.Right := Max(Left, Right);
  Result.Bottom := Min(Top, Bottom);
  Result.Top := Max(Top, Bottom);
end;

function SingleRect(P1, P2: TSinglePoint): TSingleRect;
begin
	Result := SingleRect(P1.X, P1.Y, P2.X, P2.Y);
end;

{ TGLBasePicture }

procedure TGLBasePicture.CleanupGL;
begin
	if FRC <> 0 then
  begin
  	wglMakeCurrent(0, 0);
    wglDeleteContext(FRC);
  end;
  if FPalette <> 0 then DeleteObject(FPalette);
  ReleaseDC(Handle, FDC);
end;

constructor TGLBasePicture.Create(AOwner: TComponent);
begin
  inherited;
  Color := clBlack;
  Parent := TWinControl(AOwner);
  InitVCL();
  InitGL();
  InitGL2();
end;

destructor TGLBasePicture.Destroy;
begin
	CleanupGL();
  inherited;
end;

procedure TGLBasePicture.DRAW;
begin

end;

procedure TGLBasePicture.InitDC(DC: HDC; Palette: HPalette);
var
	pfd: TPixelFormatDescriptor;
  lpPalette: PLogPalette;
  I, hHeap, nColors, nPixelFormat: Integer;
  byRedMask, byGreenMask, byBlueMask: Byte;
begin
	ZeroMemory(@pfd, SizeOf(pfd));
  pfd.nSize := SizeOf(pfd);
  pfd.nVersion := 1;
  pfd.dwFlags := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL;
  if SingleBuffer then pfd.dwFlags := pfd.dwFlags or PFD_DOUBLEBUFFER;
  if ColorIndex then pfd.iPixelType := PFD_TYPE_COLORINDEX
  else pfd.iPixelType := PFD_TYPE_RGBA;
  pfd.cColorBits := 24;
  pfd.cDepthBits := 24;
  pfd.cStencilBits := 8;
  pfd.iLayerType := PFD_MAIN_PLANE;
  nPixelFormat := ChoosePixelFormat(DC, @pfd);
  SetPixelFormat(DC, nPixelFormat, @pfd);
  if (pfd.dwFlags and PFD_NEED_PALETTE) <> 0 then
  begin
		nColors := 1 shl pfd.cColorBits;
    hHeap := GetProcessHeap();
    lpPalette := HeapAlloc(hHeap, 0,
    	SizeOf(TLogPalette) + nColors * SizeOf(TPaletteEntry));
    lpPalette^.palVersion := $300;
    lpPalette^.palNumEntries := nColors;
    byRedMask := (1 shl pfd.cRedBits) - 1;
    byGreenMask := (1 shl pfd.cGreenBits) - 1;
    byBlueMask := (1 shl pfd.cBlueBits) - 1;
    for I := 0 to nColors-1 do
    begin
    	lpPalette^.palPalEntry[I].peRed := (((i shr pfd.cRedShift) and byRedMask) * 256) div byRedMask;
			lpPalette^.palPalEntry[I].peGreen := (((i shr pfd.cGreenShift) and byGreenMask) * 256) div byGreenMask;
			lpPalette^.palPalEntry[I].peBlue := (((i shr pfd.cBlueShift) and byBlueMask) * 256) div byBlueMask;
      lpPalette^.palPalEntry[I].peFlags := 0;
   	end;
    Palette := CreatePalette(lpPalette^);
    HeapFree(hHeap, 0, lpPalette);
    if Palette <> 0 then
    begin
    	SelectPalette(DC, Palette, False);
      RealizePalette(DC);	//Map to system palette.
    end;
  end;
  SingleBuffer := True;
  ColorIndex := False;
end;

procedure TGLBasePicture.InitGL;
begin
	FDC := GetDC(Handle);
  InitDC(FDC, FPalette);
  FRC := wglCreateContext(FDC);
end;

procedure TGLBasePicture.InitGL2;
begin
	MakeCurrent();
  glEnable(GL_DEPTH);
  glEnable(GL_DEPTH_TEST);
end;

procedure TGLBasePicture.InitVCL;
begin
  Left := 0; Top := 0;
  Width := 160; Height := 90;
  Color := clBlack;
  Visible := True;
end;

procedure TGLBasePicture.MakeCurrent;
begin
	if (FDC <> 0) and (FRC <> 0) then
  	wglMakeCurrent(FDC, FRC);
end;

procedure TGLBasePicture.PageFlip;
begin
	SwapBuffers(FDC);
end;

procedure TGLBasePicture.Paint;
begin
  inherited;
	DRAW();
end;

procedure TGLBasePicture.RedoPalette;
begin
	UnrealizeObject(FPalette);
  SelectPalette(FDC, FPalette, False);
  RealizePalette(FDC);
end;

procedure TGLBasePicture.Resize;
begin
	MakeCurrent();
  glViewPort(0, 0, ClientWidth, ClientHeight);
	Invalidate();
end;

{ TGLScale }
procedure TGLScale.Change(Left, Bottom, Right, Top: Single);
begin
	FRect.Rect := SingleRect(Left, Bottom, Right, Top);
end;

procedure TGLScale.Change(Rect: TSingleRect);
begin
	FRect.Rect := Rect;
end;

procedure TGLScale.Change(P1, P2: TSinglePoint);
begin
	FRect.Rect := SingleRect(P1, P2);
end;

constructor TGLScale.Create;
begin
	inherited;
	FRect := TGLRect.Create;  
end;

destructor TGLScale.Destroy;
begin
	FreeAndNil(FRect);
  inherited;
end;

function TGLScale.GetHeight: Single;
begin
	Result := FRect.Height;
end;

function TGLScale.GetLeft: Single;
begin
	Result := FRect.Left;
end;

function TGLScale.GetRect: TSingleRect;
begin
	Result := FRect.Rect;
end;

function TGLScale.GetRight: Single;
begin
	Result := FRect.Right;
end;

function TGLScale.GetTop: Single;
begin
	Result := FRect.Top;
end;

function TGLScale.GetWidth: Single;
begin
	Result := FRect.Width;
end;

{ TGLPicture }

procedure TGLPicture.Clear;
begin
  glClearColor(
  	GetRValue(Color),
    GetGValue(Color),
    GetBValue(Color), 255);
  glClear(
  	GL_COLOR_BUFFER_BIT or
	  GL_DEPTH_BUFFER_BIT);
end;

constructor TGLPicture.Create(AOWner: TComponent);
begin
  inherited;
  FScale := TGLScale.Create;
end;

destructor TGLPicture.Destroy;
begin
  FreeAndNil(FScale);
  inherited;
end;

constructor TGLRect.Create;
begin
	inherited;
  FRect := SingleRect(0, 0, MaxInt, MaxInt);
end;

function TGLRect.GetHeight: Single;
begin
	Result := FRect.Top - FRect.Bottom;
end;

function TGLRect.GetLeft: Single;
begin
	Result := FRect.Left;
end;

function TGLRect.GetRight: Single;
begin
	Result := FRect.Right;
end;

function TGLRect.GetTop: Single;
begin
	Result := FRect.Top;
end;

function TGLRect.GetWidth: Single;
begin
	Result := FRect.Right - FRect.Left;
end;

procedure TGLPicture.Paint;
begin
end;

end.
