unit GLWindow;

interface

uses
	Windows, Classes, Forms, GL;

type
	TGLForm = class(TForm)
  private
  	FDC: HDC;
    FRC: HGLRC;
    FPalette: HPALETTE;
  public
  	constructor Create(AOwner: TComponent); override;
    destructor Destroy(); override;
    procedure InitGL();
    procedure MakeCurrent();
    procedure RedoPalette();
    procedure CleanupGL();
    procedure PageFlip();
    property DC: HDC read FDC;
    property RC: HGLRC read FRC;
    property Palette: HPALETTE read FPalette;
  end;

  TGLDeviceContext = class(TObject)
  private
  	FDC: HDC;
    FRC: HGLRC;
    FPalette: HPALETTE;
  public
		constructor Create(DC: HDC);
    destructor Destroy(); override;
    procedure InitGL();
    procedure MakeCurrent();
    procedure RedoPalette();
    procedure CleanupGL();
    procedure PageFlip();
    property DC: HDC read FDC;
    property RC: HGLRC read FRC;
    property Palette: HPALETTE read FPalette;
  end;

	procedure glInitDC(DC: HDC; Palette: HPalette);

var
	DoubleBuffer: Boolean = True;
  ColorIndex: Boolean = False;

implementation

{ Public }

procedure glInitDC(DC: HDC; Palette: HPalette);
var
	pfd: TPixelFormatDescriptor;
  lpPalette: PLogPalette;
  I, hHeap, nColors, nPixelFormat: Integer;
  byRedMask, byGreenMask, byBlueMask: Byte;
begin
	ZeroMemory(@pfd, SizeOf(pfd));
  pfd.nSize := SizeOf(pfd);
  pfd.nVersion := 1;
  pfd.dwFlags := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL;
  if DoubleBuffer then pfd.dwFlags := pfd.dwFlags or PFD_DOUBLEBUFFER;
  if ColorIndex then pfd.iPixelType := PFD_TYPE_COLORINDEX
  else pfd.iPixelType := PFD_TYPE_RGBA;
  pfd.cColorBits := 24;
  pfd.cDepthBits := 24;
  pfd.cStencilBits := 8;
  pfd.iLayerType := PFD_MAIN_PLANE;
  nPixelFormat := ChoosePixelFormat(DC, @pfd);
  SetPixelFormat(DC, nPixelFormat, @pfd);
  if (pfd.dwFlags and PFD_NEED_PALETTE) <> 0 then
  begin
		nColors := 1 shl pfd.cColorBits;
    hHeap := GetProcessHeap();
    lpPalette := HeapAlloc(hHeap, 0,
    	SizeOf(TLogPalette) + nColors * SizeOf(TPaletteEntry));
    lpPalette^.palVersion := $300;
    lpPalette^.palNumEntries := nColors;
    byRedMask := (1 shl pfd.cRedBits) - 1;
    byGreenMask := (1 shl pfd.cGreenBits) - 1;
    byBlueMask := (1 shl pfd.cBlueBits) - 1;
    for I := 0 to nColors-1 do
    begin
    	lpPalette^.palPalEntry[I].peRed := (((i shr pfd.cRedShift) and byRedMask) * 256) div byRedMask;
			lpPalette^.palPalEntry[I].peGreen := (((i shr pfd.cGreenShift) and byGreenMask) * 256) div byGreenMask;
			lpPalette^.palPalEntry[I].peBlue := (((i shr pfd.cBlueShift) and byBlueMask) * 256) div byBlueMask;
      lpPalette^.palPalEntry[I].peFlags := 0;
   	end;
    Palette := CreatePalette(lpPalette^);
    HeapFree(hHeap, 0, lpPalette);
    if Palette <> 0 then
    begin
    	SelectPalette(DC, Palette, False);
      RealizePalette(DC);	//Map to system palette.
    end;
  end;
  DoubleBuffer := True;
  ColorIndex := False;
end;

{ TGLForm }

constructor TGLForm.Create(AOwner: TComponent);
begin
  inherited;
  InitGL();
end;

destructor TGLForm.Destroy();
begin
	CleanupGL();
  inherited;
end;

procedure TGLForm.CleanupGL();
begin
	if FRC <> 0 then
  begin
  	wglMakeCurrent(0, 0);
    wglDeleteContext(FRC);
  end;
  if FPalette <> 0 then DeleteObject(FPalette);
  ReleaseDC(Handle, FDC);
end;

procedure TGLForm.MakeCurrent();
begin
	if (FDC <> 0) and (FRC <> 0) then
  	wglMakeCurrent(FDC, FRC);
end;

procedure TGLForm.InitGL();
begin
	FDC := GetDC(Handle);
  glInitDC(FDC, FPalette);
  FRC := wglCreateContext(FDC);
  MakeCurrent();
end;

procedure TGLForm.RedoPalette();
begin
	UnrealizeObject(FPalette);
  SelectPalette(FDC, FPalette, False);
  RealizePalette(FDC);
end;

procedure TGLForm.PageFlip();
begin
	SwapBuffers(FDC);
end;

{ TGLDeviceContext }

procedure TGLDeviceContext.CleanupGL;
begin
	if FRC <> 0 then
  begin
  	wglMakeCurrent(0, 0);
    wglDeleteContext(FRC);
  end;
  if FPalette <> 0 then DeleteObject(FPalette);
  DeleteDC(FDC);
end;

constructor TGLDeviceContext.Create(DC: HDC);
begin
	inherited Create();
  FDC := DC;
  InitGL();
end;

destructor TGLDeviceContext.Destroy();
begin
	CleanupGL();
  inherited;
end;

procedure TGLDeviceContext.InitGL();
begin
	glInitDC(FDC, FPalette);
  FRC := wglCreateContext(FDC);
  MakeCurrent();
end;

procedure TGLDeviceContext.MakeCurrent();
begin
	if (FDC <> 0) and (FRC <> 0) then
  	wglMakeCurrent(FDC, FRC);
end;

procedure TGLDeviceContext.PageFlip();
begin
	SwapBuffers(FDC);
end;

procedure TGLDeviceContext.RedoPalette();
begin
	UnrealizeObject(FPalette);
  SelectPalette(FDC, FPalette, False);
  RealizePalette(FDC);
end;

end.
